using System;

using Microsoft.Xna.Framework.Input;

using Syntriax.Engine.Core;
using Syntriax.Engine.Network;
using Syntriax.Engine.Systems.Input;
using Syntriax.Engine.Systems.Tween;

namespace Pong.Behaviours;

public class PaddleBehaviour(Keys Up, Keys Down, float High, float Low, float Speed) : Behaviour2D,
    IPacketListenerServer<PaddleBehaviour.PaddleKeyStatePacket>, IPacketListenerClient<PaddleBehaviour.PaddleKeyStatePacket>
{
    private Keys Up { get; } = Up;
    private Keys Down { get; } = Down;
    public float High { get; } = High;
    public float Low { get; } = Low;
    public float Speed { get; set; } = Speed;

    private bool isUpPressed = false;
    private bool isDownPressed = false;

    private IButtonInputs<Keys> inputs = null!;
    private INetworkCommunicatorClient? networkClient = null!;
    private INetworkCommunicatorServer? networkServer = null;
    private TweenManager tweenManager = null!;

    protected override void OnUpdate()
    {
        if (isUpPressed && isDownPressed)
            return;

        if (isUpPressed)
            Transform.Position += Vector2D.Up * Universe.Time.DeltaTime * Speed;
        else if (isDownPressed)
            Transform.Position += -Vector2D.Up * Universe.Time.DeltaTime * Speed;

        Transform.Position = new Vector2D(Transform.Position.X, MathF.Max(MathF.Min(Transform.Position.Y, High), Low));
    }

    protected override void OnFirstActiveFrame()
    {
        inputs = Universe.FindRequired<IButtonInputs<Keys>>();
        networkClient = Universe.Find<INetworkCommunicatorClient>();
        networkServer = Universe.Find<INetworkCommunicatorServer>();
        tweenManager = Universe.GetRequiredUniverseObject<TweenManager>();

        inputs.RegisterOnPress(Up, OnUpPressed);
        inputs.RegisterOnRelease(Up, OnUpReleased);

        inputs.RegisterOnPress(Down, OnDownPressed);
        inputs.RegisterOnRelease(Down, OnDownReleased);
    }

    protected override void OnFinalize()
    {
        inputs.UnregisterOnPress(Up, OnUpPressed);
        inputs.UnregisterOnRelease(Up, OnUpReleased);

        inputs.UnregisterOnPress(Down, OnDownPressed);
        inputs.UnregisterOnRelease(Down, OnDownReleased);
    }

    private void OnUpPressed(IButtonInputs<Keys> inputs, Keys keys) { isUpPressed = true; networkClient?.SendToServer(new PaddleKeyStatePacket(this)); }
    private void OnUpReleased(IButtonInputs<Keys> inputs, Keys keys) { isUpPressed = false; networkClient?.SendToServer(new PaddleKeyStatePacket(this)); }
    private void OnDownPressed(IButtonInputs<Keys> inputs, Keys keys) { isDownPressed = true; networkClient?.SendToServer(new PaddleKeyStatePacket(this)); }
    private void OnDownReleased(IButtonInputs<Keys> inputs, Keys keys) { isDownPressed = false; networkClient?.SendToServer(new PaddleKeyStatePacket(this)); }

    public void OnServerPacketArrived(PaddleKeyStatePacket packet, string from)
    {
        if (packet.EntityId.CompareTo(Id) != 0)
            return;

        packet.Position = Transform.Position;
        isUpPressed = packet.IsUpPressed;
        isDownPressed = packet.IsDownPressed;
        networkServer?.SendToClient("*", packet);
    }

    public void OnClientPacketArrived(PaddleKeyStatePacket packet)
    {
        isUpPressed = packet.IsUpPressed;
        isDownPressed = packet.IsDownPressed;

        if (!isUpPressed && !isDownPressed)
            Transform.Position.TweenVector2D(tweenManager, .05f, packet.Position, x => Transform.Position = x);
        else
            Transform.Position = packet.Position;
    }

    public class PaddleKeyStatePacket : IEntityNetworkPacket
    {
        public string EntityId { get; set; } = default!;
        public bool IsUpPressed { get; set; } = default!;
        public bool IsDownPressed { get; set; } = default!;
        public Vector2D Position { get; set; } = default!;
        public long TimeTicker { get; set; } = default!;

        public PaddleKeyStatePacket() { }
        public PaddleKeyStatePacket(PaddleBehaviour paddleBehaviour)
        {
            EntityId = paddleBehaviour.Id;
            IsUpPressed = paddleBehaviour.isUpPressed;
            IsDownPressed = paddleBehaviour.isDownPressed;
            Position = paddleBehaviour.Transform.Position;
            TimeTicker = DateTime.UtcNow.Ticks;
        }
    }
}
